@@ -1,4 +1,4 @@
-from sys import path
+from sys import path, version_info
 from os.path import sep
 path.insert(1, path[0]+sep+'codec'+sep+'ber')
 import ber.suite
@@ -7,12 +7,14 @@ import cer.suite
 path.insert(1, path[0]+sep+'codec'+sep+'der')
 import der.suite
 from pyasn1.error import PyAsn1Error
-try:
+if version_info[0:2] < (2, 7) or \
+   version_info[0:2] in ( (3, 0), (3, 1) ):
+    try:
+        import unittest2 as unittest
+    except ImportError:
+        import unittest
+else:
     import unittest
-except ImportError:
-    raise PyAsn1Error(
-        'PyUnit package\'s missing. See http://pyunit.sourceforge.net/'
-        )
 
 suite = unittest.TestSuite()
 for m in (
